#!/usr/bin/env python3
#
#   test_02_ipmi.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.Ipmi() class.
#
import subprocess
from typing import Any, List
from configparser import ConfigParser
import pytest
from mock import MagicMock, call
from pytest_mock import MockerFixture
from smfc import Log, Ipmi
from .test_00_data import TestData

class TestIpmi:
    """Unit test class for smfc.Ipmi() class"""

    @pytest.mark.parametrize("mode_delay, level_delay, remote_pars, sudo, error", [
        (10, 2, '',                                          False, 'Ipmi.__init__() 1'),
        (2, 10, '-I lanplus -U ADMIN -P ADMIN -H 127.0.0.1', True,  'Ipmi.__init__() 2')
    ])
    def test_init_p1(self, mocker: MockerFixture, mode_delay: int, level_delay: int,
                     remote_pars: str, sudo: bool, error: str) -> None:
        """Positive unit test function for Ipmi.__init__() method. It contains the following steps:
            - create a shell script for IPMI command
            - mock print() function
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if the class attributes contain different values that were passed to __init__
            - ASSERT: if the mocked print function was called wrong number of times
        """
        my_td = TestData()
        command = my_td.create_command_file()
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_ipmi_exec = MagicMock()
        mock_ipmi_exec.return_value = subprocess.CompletedProcess([], returncode=0)
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: command,
            Ipmi.CV_IPMI_FAN_MODE_DELAY: str(mode_delay),
            Ipmi.CV_IPMI_FAN_LEVEL_DELAY: str(level_delay),
            Ipmi.CV_IPMI_REMOTE_PARAMETERS: remote_pars
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi(my_log, my_config, sudo)
        assert my_ipmi.command == command, error
        assert my_ipmi.fan_mode_delay == mode_delay, error
        assert my_ipmi.fan_level_delay == level_delay, error
        assert my_ipmi.remote_parameters == remote_pars, error
        assert mock_print.call_count == 5, error  # Ipmi-5
        assert my_ipmi.sudo == sudo, error
        del my_td

    @pytest.mark.parametrize("cmd_exists, mode_delay, level_delay, remote_pars, exception, error", [
        (True, -1,  2,  None,            ValueError,         'Ipmi.__init__() 3'),
        (True, 10, -2,  '-I lanplus',    ValueError,         'Ipmi.__init__() 4'),
        (False, 1,  1,  '',              FileNotFoundError,  'Ipmi.__init__() 5'),
        (False, 1,  1,  '-I lanplus',    RuntimeError,       'Ipmi.__init__() 6')
    ])
    def test_init_n1(self, mocker: MockerFixture, cmd_exists: bool, mode_delay: int, level_delay: int,
                     remote_pars: str, exception: Any, error: str) -> None:
        """Negative unit test for Ipmi.__init__() method. It contains the following steps:
            - create a shell script depending on `cmd_exists` flag
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if the specified exception was not raised during __init__
            - delete all instances
        """

        #pylint: disable=W0613
        def mocked_ipmi_exec(self, args: List[str]) -> subprocess.CompletedProcess:
            if exception is not ValueError:
                raise exception
            return subprocess.CompletedProcess([], returncode=0)
        #pylint: enable=W0613

        my_td = TestData()
        command = my_td.create_command_file()
        if not cmd_exists:
            my_td.delete_file(command)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker.patch('smfc.Ipmi._exec_ipmitool', mocked_ipmi_exec)
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: command,
            Ipmi.CV_IPMI_FAN_MODE_DELAY: str(mode_delay),
            Ipmi.CV_IPMI_FAN_LEVEL_DELAY: str(level_delay)
        }
        if remote_pars is not None:
            my_config.set(Ipmi.CS_IPMI, Ipmi.CV_IPMI_REMOTE_PARAMETERS, remote_pars)
        my_log = Log(Log.LOG_ERROR, Log.LOG_STDOUT)
        with pytest.raises(Exception) as cm:
            Ipmi(my_log, my_config, False)
        assert cm.type == exception, error
        del my_td

    # pylint: disable=duplicate-code, protected-access
    @pytest.mark.parametrize("args, remote_args, sudo, error", [
        (['1', '2', '3', '4', '5'], '',             False, 'Ipmi.exec() 1'),
        (['1', '2', '3', '4', '5'], '',             True,  'Ipmi.exec() 2'),
        (['1', '2', '3', '4', '5'], '-I lanplus',   False, 'Ipmi.exec() 3'),
        (['1', '2', '3', '4', '5'], '-I lanplus',   True,  'Ipmi.exec() 4'),
        ([],                        '',             False, 'Ipmi.exec() 5'),
        ([],                        '',             True,  'Ipmi.exec() 6'),
        ([],                        '-I lanplus',   False, 'Ipmi.exec() 7'),
        ([],                        '-I lanplus',   True,  'Ipmi.exec() 8')
    ])
    def test_exec_ipmitool_p(self, mocker: MockerFixture, args: List[str], remote_args: str, sudo:bool,
                             error: str) -> None:
        """Positive unit test for Ipmi.exec() method. It contains the following steps:
            - mock print(), subprocess.run() functions
            - create an Ipmi classes
            - Call Ipmi._exec_ipmitool() method
            - ASSERT: if it was called with different parameters from expected argument list
            - ASSERT: if it was called with different times from expected value
        """
        expected: List[str]     # Expected argument list.

        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_subprocess_run = MagicMock()
        mocker.patch('subprocess.run', mock_subprocess_run)
        mock_subprocess_run.return_value = subprocess.CompletedProcess([], returncode=0)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.command = 'usr/bin/ipmitool'
        my_ipmi.remote_parameters = remote_args
        my_ipmi.sudo = sudo
        my_ipmi._exec_ipmitool(args)
        expected = []
        if sudo:
            expected.append('sudo')
        expected.append(my_ipmi.command)
        if remote_args:
            expected.extend(remote_args.split())
        expected.extend(args)
        mock_subprocess_run.assert_called_with(expected, check=False, capture_output=True, text=True)
        assert mock_subprocess_run.call_count == 1, error

    @pytest.mark.parametrize("ipmi_command, sudo, rc, exception, error", [
        # The real subprocess.run() executed (without sudo)
        ('/nonexistent/command', False, 0, FileNotFoundError, 'Ipmi.exec() 9'),
        # The mocked subprocess.run() executed and returns non-zero return code
        ('',                     True,  1, RuntimeError,      'Ipmi.exec() 10'),
        ('',                     False, 1, RuntimeError,      'Ipmi.exec() 10')
    ])
    def test_exec_ipmitool_n(self, mocker: MockerFixture, ipmi_command, sudo:bool, rc: int, exception: Any,
                             error: str) -> None:
        """Negative unit test for Ipmi.exec() method. It contains the following steps:
            - create a shell script for IPMI command
            - mock print(), subprocess.run() functions
            - initialize a Config, Log, Ipmi classes
            - Call Ipmi._exec_ipmitool() method
            - ASSERT: if the expected assertion was not raised
        """
        err: List[str] = [
            'sudo: ipmi command not found',
            'ipmitool: error while loading shared libraries'
        ]
        # If we need to mock for the return code.
        if rc:
            mock_subprocess_run = MagicMock()
            mocker.patch('subprocess.run', mock_subprocess_run)
            mock_subprocess_run.return_value = subprocess.CompletedProcess([], returncode=rc,
                                                                            stderr=err[0] if sudo else err[1])
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.command = ipmi_command
        my_ipmi.remote_parameters = ''
        my_ipmi.sudo = sudo
        with pytest.raises(Exception) as cm:
            my_ipmi._exec_ipmitool(['1', '2', '3'])
        assert cm.type == exception, error
    # pylint: enable=duplicate-code, protected-access

    @pytest.mark.parametrize("expected_mode, error", [
        (Ipmi.STANDARD_MODE,    'Ipmi.get_fan_mode() 1'),
        (Ipmi.FULL_MODE,        'Ipmi.get_fan_mode() 2'),
        (Ipmi.OPTIMAL_MODE,     'Ipmi.get_fan_mode() 3'),
        (Ipmi.HEAVY_IO_MODE,    'Ipmi.get_fan_mode() 4')
    ])
    def test_get_fan_mode_p1(self, mocker:MockerFixture, expected_mode: int, error: str) -> None:
        """Positive unit test for Ipmi.get_fan_mode() method. It contains the following steps:
            - create a shell script with an expected output
            - mock print() function
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if the get_fan_mode() returns different value from the expected one
            - delete all instances
        """
        my_td = TestData()
        command = my_td.create_command_file(f'echo " {expected_mode:02}"')
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: command
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi(my_log, my_config, False)
        assert my_ipmi.get_fan_mode() == expected_mode, error
        del my_td

    @pytest.mark.parametrize("value, exception, error", [
        ('NA', ValueError, 'Ipmi.get_fan_mode() 5'),
        ('',   ValueError, 'Ipmi.get_fan_mode() 6')
    ])
    def test_get_fan_mode_n1(self, value: str, exception: Any, error: str) -> None:
        """Negative unit test for Ipmi.get_fan_mode() method. It contains the following steps:
            - create a shell script providing invalid value
            - initialize a Config, Log, Ipmi classes
            - call get_fan_mode() function
            - ASSERT: if the no ValueError exception raised (other exceptions are tested in .exec() method)
            - delete all instances
        """
        my_td = TestData()
        command = my_td.create_command_file('echo " ' + value + '"')
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: command
        }
        my_log = Log(Log.LOG_ERROR, Log.LOG_STDOUT)
        my_ipmi = Ipmi(my_log, my_config, False)
        with pytest.raises(Exception) as cm:
            my_ipmi.get_fan_mode()
        assert cm.type == exception, error
        del my_td

    @pytest.mark.parametrize("fm, fms, error", [
        (Ipmi.STANDARD_MODE, 'STANDARD', 'Ipmi.get_fan_mode_name() 1'),
        (Ipmi.FULL_MODE,     'FULL',     'Ipmi.get_fan_mode_name() 2'),
        (Ipmi.OPTIMAL_MODE,  'OPTIMAL',  'Ipmi.get_fan_mode_name() 3'),
        (Ipmi.PUE_MODE,      'PUE',      'Ipmi.get_fan_mode_name() 4'),
        (Ipmi.HEAVY_IO_MODE, 'HEAVY IO', 'Ipmi.get_fan_mode_name() 5'),
        (100,                'UNKNOWN',  'Ipmi.get_fan_mode_name() 6')
    ])
    def test_get_fan_mode_name(self, fm: int, fms: str, error: str) -> None:
        """Positive unit test for Ipmi.get_fan_mode_name() method. It contains the following steps:
            - create a shell script for ipmitool substitution
            - mock print() function
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if the get_fan_mode_name() returns with a different string from the expected one
            - delete all instances
        """
        assert Ipmi.get_fan_mode_name(fm) == fms, error

    @pytest.mark.parametrize("fan_mode, error", [
        (Ipmi.STANDARD_MODE, 'Ipmi.set_fan_mode() 1'),
        (Ipmi.FULL_MODE,     'Ipmi.set_fan_mode() 2'),
        (Ipmi.OPTIMAL_MODE,  'Ipmi.set_fan_mode() 3'),
        (Ipmi.PUE_MODE,      'Ipmi.set_fan_mode() 4'),
        (Ipmi.HEAVY_IO_MODE, 'Ipmi.set_fan_mode() 5')
    ])
    def test_set_fan_mode_p1(self, mocker:MockerFixture, fan_mode: int, error: str) -> None:
        """Positive unit test for Ipmi.set_fan_mode() method. It contains the following steps:
            - mock Ipmi.exec() and time.sleep() functions
            - create an empty Ipmi class
            - ASSERT: if set_fan_mode() calls Ipmi.exec() and time.sleep() other parameters from expected
            - ASSERT: if set_fan_mode() calls Ipmi.exec() and time.sleep() more from expected times
        """
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_mode_delay = 0
        mock_ipmi_exec = MagicMock()
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        mock_time_sleep = MagicMock()
        mocker.patch('time.sleep', mock_time_sleep)
        my_ipmi.set_fan_mode(fan_mode)
        mock_ipmi_exec.assert_called_with(['raw', '0x30', '0x45', '0x01', f'0x{fan_mode:02x}'])
        assert mock_ipmi_exec.call_count == 1, error
        mock_time_sleep.assert_called_with(my_ipmi.fan_mode_delay)
        assert mock_time_sleep.call_count == 1, error

    @pytest.mark.parametrize("fan_mode, exception, error", [
        (-1,  ValueError, 'Ipmi.set_fan_mode() 6'),
        (100, ValueError, 'Ipmi.set_fan_mode() 7')
    ])
    def test_set_fan_mode_n1(self, mocker: MockerFixture, fan_mode: int, exception: Any, error: str) -> None:
        """Negative unit test for Ipmi.set_fan_mode(). It contains the following steps:
            - mock Ipmi.exec() function
            - create an empty Ipmi class
            - ASSERT: if set_fan_mode() did not raise ValueError exception in case of invalid parameters
              (other potential exceptions are tested elsewhere)
        """
        mock_ipmi_exec = MagicMock()
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_mode_delay = 0
        my_ipmi.sudo = False
        with pytest.raises(ValueError) as cm:
            my_ipmi.set_fan_mode(fan_mode)
        assert cm.type == exception, error

    @pytest.mark.parametrize("zone, level, error", [
        (0, 0,        'Ipmi.set_fan_level() 1'),
        (0, 50,       'Ipmi.set_fan_level() 2'),
        (0, 100,      'Ipmi.set_fan_level() 3'),
        (1, 0,        'Ipmi.set_fan_level() 4'),
        (1, 50,       'Ipmi.set_fan_level() 5'),
        (1, 100,      'Ipmi.set_fan_level() 6')
    ])
    def test_set_fan_level_p1(self, mocker:MockerFixture, zone: int, level: int, error: str) -> None:
        """Positive unit test function. It contains the following steps:
            - mock print(), subprocess.run() functions
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if set_fan_level() calls subprocess.run() command with other parameters than expected
            - delete the instances
        """
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_level_delay = 0
        mock_ipmi_exec = MagicMock()
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        mock_time_sleep = MagicMock()
        mocker.patch('time.sleep', mock_time_sleep)
        my_ipmi.set_fan_level(zone, level)
        mock_ipmi_exec.assert_called_with(['raw', '0x30', '0x70', '0x66', '0x01', f'0x{zone:02x}', f'0x{level:02x}'])
        assert mock_ipmi_exec.call_count == 1, error
        mock_time_sleep.assert_called_with(my_ipmi.fan_level_delay)
        assert mock_time_sleep.call_count == 1, error

    @pytest.mark.parametrize("zone, level, error", [
        (Ipmi.CPU_ZONE, -1,  'Ipmi.set_fan_level() 7'),
        (Ipmi.CPU_ZONE, 101, 'Ipmi.set_fan_level() 8'),
        (-1,            50,  'Ipmi.set_fan_level() 9'),
        (101,           50,  'Ipmi.set_fan_level() 10')
    ])
    def test_set_fan_level_n1(self, mocker: MockerFixture, zone: int, level: int, error: str) -> None:
        """Negative unit test for Ipmi.set_fan_level() method. It contains the following steps:
            - mock Ipmi.exec() function
            - create an empty Ipmi class
            - ASSERT: if set_fan_level() does not raise ValueError exception in case of invalid parameter
              (other exceptions are tested elsewhere)
        """
        mock_ipmi_exec = MagicMock()
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_level_delay = 0
        my_ipmi.sudo = False
        with pytest.raises(ValueError) as cm:
            my_ipmi.set_fan_level(zone, level)
        assert cm.type is ValueError, error

    @pytest.mark.parametrize("zones, level, error", [
        ([0],       0,      'Ipmi.set_multiple_fan_levels() 1'),
        ([0, 1],    50,     'Ipmi.set_multiple_fan_levels() 2'),
        ([0, 1, 2], 100,    'Ipmi.set_multiple_fan_levels() 3')
    ])
    def test_set_multiple_fan_levels_p1(self, mocker:MockerFixture, zones: List[int], level: int, error: str) -> None:
        """Positive unit test function. It contains the following steps:
            - mock print(), subprocess.run() functions
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if set_multiple_fan_levels() calls subprocess.run() command with other parameters than expected
            - delete the instances
        """
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_level_delay = 0
        mock_ipmi_exec = MagicMock()
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        mock_time_sleep = MagicMock()
        mocker.patch('time.sleep', mock_time_sleep)
        my_ipmi.set_multiple_fan_levels(zones, level)
        calls=[]
        for z in zones:
            calls.append(call(['raw', '0x30', '0x70', '0x66', '0x01', f'0x{z:02x}', f'0x{level:02x}']))
        mock_ipmi_exec.assert_has_calls(calls)
        assert mock_ipmi_exec.call_count == len(zones), error
        mock_time_sleep.assert_called_with(my_ipmi.fan_level_delay)
        assert mock_time_sleep.call_count == 1, error

    @pytest.mark.parametrize("zones, level, error", [
        ([0],       -1,     'Ipmi.set_multiple_fan_levels() 4'),
        ([0],       101,    'Ipmi.set_multiple_fan_levels() 5'),
        ([-1],      50,     'Ipmi.set_multiple_fan_levels() 6'),
        ([101],     50,     'Ipmi.set_multiple_fan_levels() 7'),
        ([0, -1],   50,     'Ipmi.set_multiple_fan_levels() 8'),
        ([101, 0],  50,     'Ipmi.set_multiple_fan_levels() 9')
    ])
    def test_set_multiple_fan_levels_n1(self, mocker: MockerFixture, zones: List[int], level: int, error: str) -> None:
        """Negative unit test for Ipmi.set_fan_level() method. It contains the following steps:
            - mock Ipmi.exec() function
            - create an empty Ipmi class
            - ASSERT: if set_multiple_fan_levels() does not raise ValueError exception in case of invalid parameters
              (other exceptions are tested elsewhere)
        """
        mock_ipmi_exec = MagicMock()
        mocker.patch('smfc.Ipmi._exec_ipmitool', mock_ipmi_exec)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_level_delay = 0
        my_ipmi.sudo = False
        with pytest.raises(ValueError) as cm:
            my_ipmi.set_multiple_fan_levels(zones, level)
        assert cm.type is ValueError, error

    @pytest.mark.parametrize("zone, expected_level, error", [
        (Ipmi.CPU_ZONE, 0,   'Ipmi.get_fan_level() 1'),
        (Ipmi.CPU_ZONE, 50,  'Ipmi.get_fan_level() 2'),
        (Ipmi.CPU_ZONE, 100, 'Ipmi.get_fan_level() 3'),
        (Ipmi.HD_ZONE,  0,   'Ipmi.get_fan_level() 4'),
        (Ipmi.HD_ZONE,  50,  'Ipmi.get_fan_level() 5'),
        (Ipmi.HD_ZONE,  100, 'Ipmi.get_fan_level() 6')
    ])
    def test_get_fan_level_p1(self, mocker:MockerFixture, zone: int, expected_level: int, error: str) -> None:
        """Positive unit test for Ipmi.get_fan_level() method. It contains the following steps:
            - create a shell script with the expected output
            - mock print() function
            - initialize a Config, Log, Ipmi classes
            - ASSERT: if the get_fan_level() returns different from the expected value
            - delete all instances
        """
        my_td = TestData()
        command = my_td.create_command_file(f'echo " {expected_level:x}"')
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: command
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi(my_log, my_config, False)
        assert my_ipmi.get_fan_level(zone) == expected_level, error
        del my_td

    @pytest.mark.parametrize("zone, level, error", [
        (Ipmi.CPU_ZONE, 'NA', 'Ipmi.get_fan_level() 7'),
        (Ipmi.CPU_ZONE, '',   'Ipmi.get_fan_level() 8'),
        (Ipmi.HD_ZONE,  'NA', 'Ipmi.get_fan_level() 9'),
        (Ipmi.HD_ZONE,  '',   'Ipmi.get_fan_level() 10'),
        (-1,            'NA', 'Ipmi.get_fan_level() 11'),
        (200,           '',   'Ipmi.get_fan_level() 12')
    ])
    def test_get_fan_level_n1(self, zone: int, level: str, error: str) -> None:
        """Negative unit test for Ipmi.get_fan_mode() method. It contains the following steps:
            - create a shell script providing invalid value
            - initialize a Config, Log, Ipmi classes
            - call get_fan_level() function
            - ASSERT: if no ValueError exception raised (other exceptions are tested in .exec() method)
            - delete all instances
        """
        my_td = TestData()
        command = my_td.create_command_file('echo " '+level+'"')
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: command
        }
        my_log = Log(Log.LOG_ERROR, Log.LOG_STDOUT)
        my_ipmi = Ipmi(my_log, my_config, False)
        with pytest.raises(Exception) as cm:
            my_ipmi.get_fan_level(zone)
        assert cm.type is ValueError, error
        del my_td

    @pytest.mark.parametrize("exception, error", [
        (RuntimeError,      'Ipmi exceptions 1'),
        (FileNotFoundError, 'Ipmi exceptions 2')
    ])
    def test_exceptions(self, mocker:MockerFixture, exception: Any, error: str) -> None:
        """Negative unit test for Ipmi.get_fan_mode(), Ipmi.set_fan_mode(), Ipmi.set_fan_level(),
           Ipmi.get_fan_level() methods. It contains the following steps:
            - create a shell script providing invalid value
            - initialize an empty Ipmi class
            - call all functions above
            - ASSERT: if the expected exception was not raised
        """

        def mocked_ipmi_exec(self, args: List[str]) -> subprocess.CompletedProcess:
            raise exception

        mocker.patch('smfc.Ipmi._exec_ipmitool', mocked_ipmi_exec)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_ipmi.fan_mode_delay = 0
        my_ipmi.fan_level_delay = 0
        my_ipmi.sudo = False

        with pytest.raises(Exception) as cm:
            my_ipmi.get_fan_mode()
        assert cm.type == exception, error

        with pytest.raises(Exception) as cm:
            my_ipmi.set_fan_mode(Ipmi.FULL_MODE)
        assert cm.type == exception, error

        with pytest.raises(Exception) as cm:
            my_ipmi.set_fan_level(Ipmi.CPU_ZONE, 50)
        assert cm.type == exception, error

        with pytest.raises(Exception) as cm:
            my_ipmi.set_multiple_fan_levels([0], 50)
        assert cm.type == exception, error

        with pytest.raises(Exception) as cm:
            my_ipmi.get_fan_level(Ipmi.CPU_ZONE)
        assert cm.type == exception, error


# End.



#!/usr/bin/env python3
#
#   test_05_hdzone.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.HdZone() class.
#
import random
import subprocess
import os
import time
from configparser import ConfigParser
from typing import List, Any
import pytest
import pyudev
from mock import MagicMock
from pytest_mock import MockerFixture
from smfc import Log, Ipmi, FanController, HdZone
from .test_00_data import TestData, MockDevices, factory_mockdevice

class TestHdZone:
    """Unit test class for smfc.HdZone() class"""

    @pytest.mark.parametrize(
        "count, ipmi_zone, temp_calc, steps, sensitivity, polling, min_temp, max_temp, min_level, max_level, sb_limit, "
        "sudo, error", [
        # Test valid parameters (hd=1 case is not tested because it turns off standby guard).
        (1, '0', FanController.CALC_MIN, 4, 2, 2, 32, 48, 35, 100, 2, True,  'HdZone.__init__() 1'),
        (2, '1', FanController.CALC_AVG, 4, 2, 2, 32, 48, 35, 100, 2, False, 'HdZone.__init__() 2'),
        (4, '2', FanController.CALC_AVG, 4, 2, 2, 32, 48, 35, 100, 4, True,  'HdZone.__init__() 3'),
        (8, '3', FanController.CALC_MAX, 4, 2, 2, 32, 48, 35, 100, 6, False, 'HdZone.__init__() 4')
    ])
    def test_init_p1(self, mocker: MockerFixture, count: int, ipmi_zone: str, temp_calc: int, steps: int,
                     sensitivity: float, polling: float, min_temp: float, max_temp: float, min_level: int,
                     max_level: int, sb_limit: int, sudo: bool, error: str):
        """Positive unit test for HdZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Devices.from_device_file(), pyudev.Device, smfc.FanController.get_hwmon_path()
            - initialize a Config, Log, Context, Ipmi, and HdZone classes
            - ASSERT: if the HdZone class attributes are different from values passed to __init__
        """
        my_td = TestData()
        cmd_smart = my_td.create_command_file('echo "ACTIVE"')
        my_td.create_hd_data(count)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker.patch.object(pyudev.Device, '__new__', new_callable=factory_mockdevice)
        mocker.patch('pyudev.Devices.from_device_file', MockDevices.from_device_file)
        mock_ipmi_exec = MagicMock()
        mock_ipmi_exec.return_value = subprocess.CompletedProcess([], returncode=0)
        mocker.patch('smfc.HdZone._exec_smartctl', mock_ipmi_exec)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.hd_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: '1',
            HdZone.CV_HD_IPMI_ZONE: ipmi_zone,
            HdZone.CV_HD_ZONE_TEMP_CALC: str(temp_calc),
            HdZone.CV_HD_ZONE_STEPS: str(steps),
            HdZone.CV_HD_ZONE_SENSITIVITY: str(sensitivity),
            HdZone.CV_HD_ZONE_POLLING: str(polling),
            HdZone.CV_HD_ZONE_MIN_TEMP: str(min_temp),
            HdZone.CV_HD_ZONE_MAX_TEMP: str(max_temp),
            HdZone.CV_HD_ZONE_MIN_LEVEL: str(min_level),
            HdZone.CV_HD_ZONE_MAX_LEVEL: str(max_level),
            HdZone.CV_HD_ZONE_HD_NAMES: my_td.hd_names,
            HdZone.CV_HD_ZONE_SMARTCTL_PATH: cmd_smart,
            HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED: '1',
            HdZone.CV_HD_ZONE_STANDBY_HD_LIMIT: str(sb_limit)
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        my_hdzone = HdZone(my_log, my_udevc, my_ipmi, my_config, sudo)
        assert my_hdzone.ipmi_zone == [int(s) for s in ipmi_zone.split(',' if ',' in ipmi_zone else ' ')], error
        assert my_hdzone.name == HdZone.CS_HD_ZONE, error
        assert my_hdzone.count == count, error
        assert my_hdzone.sudo == sudo, error
        assert my_hdzone.temp_calc == temp_calc, error
        assert my_hdzone.steps == steps, error
        assert my_hdzone.sensitivity == sensitivity, error
        assert my_hdzone.polling == polling, error
        assert my_hdzone.min_temp == min_temp, error
        assert my_hdzone.max_temp == max_temp, error
        assert my_hdzone.min_level == min_level, error
        assert my_hdzone.max_level == max_level, error
        assert my_hdzone.hd_device_names == my_td.hd_name_list, error
        assert my_hdzone.smartctl_path == cmd_smart, error
        assert my_hdzone.hwmon_path == my_td.hd_files, error
        if count > 1:
            assert my_hdzone.standby_hd_limit == sb_limit, error
            assert my_hdzone.standby_guard_enabled is True, error
        del my_td

    @pytest.mark.parametrize("error", [
        'HdZone.__init__() 5'
    ])
    def test_init_p2(self, mocker: MockerFixture, error: str):
        """Positive unit test for HdZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Devices.from_device_file(), pyudev.Device, smfc.FanController.get_hwmon_path()
            - initialize a Config, Log, Context, Ipmi, and HdZone classes
            - ASSERT: if the HdZone class attributes are different from the default configuration values
        """
        my_td = TestData()
        count = 4
        my_td.create_hd_data(count)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker.patch.object(pyudev.Device, '__new__', new_callable=factory_mockdevice)
        mocker.patch('pyudev.Devices.from_device_file', MockDevices.from_device_file)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.hd_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: '1',
            HdZone.CV_HD_ZONE_HD_NAMES: my_td.hd_names
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        my_hdzone = HdZone(my_log, my_udevc, my_ipmi, my_config, False)
        assert my_hdzone.log == my_log, error
        assert my_hdzone.ipmi == my_ipmi
        assert my_hdzone.ipmi_zone == [Ipmi.HD_ZONE], error
        assert my_hdzone.name == HdZone.CS_HD_ZONE, error
        assert my_hdzone.count == count, error
        assert my_hdzone.sudo is False
        assert my_hdzone.temp_calc == FanController.CALC_AVG, error
        assert my_hdzone.steps == 4, error
        assert my_hdzone.sensitivity == 2, error
        assert my_hdzone.polling == 10, error
        assert my_hdzone.min_temp == 32, error
        assert my_hdzone.max_temp == 46, error
        assert my_hdzone.min_level == 35, error
        assert my_hdzone.max_level == 100, error
        assert my_hdzone.hd_device_names == my_td.hd_name_list, error
        assert my_hdzone.smartctl_path == '/usr/sbin/smartctl', error
        assert my_hdzone.hwmon_path == my_td.hd_files, error
        del my_td

    @pytest.mark.parametrize(
        "count, temp_calc, steps, sensitivity, polling, min_temp, max_temp, min_level, max_level, sb_limit, error", [
        # hd_names= not specified (count = 0)
        (0,     FanController.CALC_MIN, 4, 2, 2, 32, 48, 35, 100, 2,  'HdZone.__init__() 6'),
        # standby_hd_limit < 0
        (2,     FanController.CALC_MIN, 4, 2, 2, 32, 48, 35, 100, -1, 'HdZone.__init__() 7'),
        # standby_hd_limit > count
        (2,     FanController.CALC_MIN, 4, 2, 2, 32, 48, 35, 100, 4,  'HdZone.__init__() 8'),
        # Invalid device name (count == 100)
        (100,   FanController.CALC_MIN, 4, 2, 2, 32, 48, 35, 100, 4,  'HdZone.__init__() 9')
    ])
    def test_init_n1(self, mocker: MockerFixture, count: int, temp_calc: int, steps: int, sensitivity: float,
                     polling: float, min_temp: float, max_temp: float, min_level: int, max_level: int, sb_limit: int,
                     error: str):
        """Negative unit test for HdZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Devices.from_device_file(), pyudev.Device, smfc.FanController.get_hwmon_path()
            - initialize a Config, Log, Ipmi, and HdZone classes
            - ASSERT: if no assertion is raised for invalid values at initialization
        """
        my_td = TestData()
        cmd_smart = my_td.create_command_file('echo "ACTIVE"')
        if count == 100:
            my_td.create_hd_data(1)
            my_td.hd_names='raise\n'
        else:
            my_td.create_hd_data(count)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker.patch.object(pyudev.Device, '__new__', new_callable=factory_mockdevice)
        mocker.patch('pyudev.Devices.from_device_file', MockDevices.from_device_file)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.hd_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: '1',
            HdZone.CV_HD_ZONE_TEMP_CALC: str(temp_calc),
            HdZone.CV_HD_ZONE_STEPS: str(steps),
            HdZone.CV_HD_ZONE_SENSITIVITY: str(sensitivity),
            HdZone.CV_HD_ZONE_POLLING: str(polling),
            HdZone.CV_HD_ZONE_MIN_TEMP: str(min_temp),
            HdZone.CV_HD_ZONE_MAX_TEMP: str(max_temp),
            HdZone.CV_HD_ZONE_MIN_LEVEL: str(min_level),
            HdZone.CV_HD_ZONE_MAX_LEVEL: str(max_level),
            HdZone.CV_HD_ZONE_HD_NAMES: my_td.hd_names,
            HdZone.CV_HD_ZONE_SMARTCTL_PATH: cmd_smart,
            HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED: '1',
            HdZone.CV_HD_ZONE_STANDBY_HD_LIMIT: str(sb_limit)
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        with pytest.raises(Exception) as cm:
            HdZone(my_log, my_udevc, my_ipmi, my_config, False)
        assert cm.type is ValueError, error
        del my_td

    #pylint: disable=protected-access
    @pytest.mark.parametrize("args, sudo, error", [
        (['-a', '/dev/sda'],                    True,  'HdZone._exec_smartctl() 1'),
        (['-a', '/dev/sda'],                    False, 'HdZone._exec_smartctl() 2'),
        (['-i', '-n', 'standby', '/dev/sda'],   True,  'HdZone._exec_smartctl() 3'),
        (['-i', '-n', 'standby', '/dev/sda'],   False, 'HdZone._exec_smartctl() 4'),
        (['-s', '/dev/sda'],                    True,  'HdZone._exec_smartctl() 5'),
        (['-s', '/dev/sda'],                    False, 'HdZone._exec_smartctl() 6')
    ])
    def test_exec_smartctl_p(self, mocker: MockerFixture, args:List[str], sudo: bool, error: str):
        """Positive unit test for HdZone._exec_smartctl() method. It contains the following steps:
            - mock subprocess.run() function
            - initialize an empty HdZone class
            - call HdZone._exec_smartctl() method
            - ASSERT: if subprocess.run() called different from specified argument list
        """
        expected_args: List[str]

        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.smartctl_path = 'smartctl'
        my_hdzone.sudo = sudo
        mock_subprocess_run = MagicMock()
        mock_subprocess_run.return_value=subprocess.CompletedProcess([], returncode=0, stdout='', stderr='')
        mocker.patch('subprocess.run', mock_subprocess_run)
        my_hdzone._exec_smartctl(args)
        expected_args = []
        if sudo:
            expected_args.append('sudo')
        expected_args.append(my_hdzone.smartctl_path)
        expected_args.extend(args)
        mock_subprocess_run.assert_called_with(expected_args, capture_output=True, check=False, text=True)
        assert mock_subprocess_run.call_count == 1, error

    #pylint: disable=R0801
    @pytest.mark.parametrize("smartctl_command, sudo, rc, exception, error", [
        # The real subprocess.run() executed (without sudo)
        ('/nonexistent/command', False, 0, FileNotFoundError, 'HdZone._exec_smartctl() 7'),
        # The mocked subprocess.run() executed and returns non-zero return code
        ('',                     True,  1, RuntimeError,      'HdZone._exec_smartctl() 8')
    ])
    def test_exec_smartctl_n(self, mocker: MockerFixture, smartctl_command, sudo: bool, rc: int, exception: Any,
                             error: str):
        """Negative unit test for HdZone._exec_smartctl() method. It contains the following steps:
            - mock subprocess.run() function if needed
            - initialize an empty HdZone class
            - call HdZone._exec_smartctl() method
            - ASSERT: if no assertion was raised
        """
        if rc:
            mock_subprocess_run = MagicMock()
            mocker.patch('subprocess.run', mock_subprocess_run)
            mock_subprocess_run.return_value = subprocess.CompletedProcess([], returncode=rc,
                                                                           stderr='sudo: smartctl: command not found')
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.smartctl_path = smartctl_command
        my_hdzone.sudo = sudo
        with pytest.raises(Exception) as cm:
            my_hdzone._exec_smartctl(['-a', '/dev/sda'])
        assert cm.type == exception, error
    # pylint: enable=R0801

    @pytest.mark.parametrize("count, temperatures, error", [
        (1, [32],                               'HdZone._get_nth_temp() 1'),
        (2, [33, 34],                           'HdZone._get_nth_temp() 2'),
        (4, [33, 34, 35, 38],                   'HdZone._get_nth_temp() 3'),
        (8, [33, 34, 35, 38, 36, 37, 31, 30],   'HdZone._get_nth_temp() 4')
    ])
    def test_get_ntf_temp_p1(self, mocker: MockerFixture, count: int, temperatures: List[float], error: str):
        """Positive unit test for HdZone._get_nth_temp() method. It contains the following steps:
            - mock print() function
            - initialize an empty HdZone class
            - ASSERT: if the read temperature (from HWMON) is different from the expected value
        """
        my_td = TestData()
        my_td.create_hd_data(count, temperatures)
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.hwmon_path = my_td.hd_files
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        for i in range(count):
            temp = my_hdzone._get_nth_temp(i)
            assert temp == temperatures[i], error
        del my_td

    @pytest.mark.parametrize("count, temperatures, error", [
        (1, [32],                               'HdZone._get_nth_temp() 5'),
        (2, [33, 34],                           'HdZone._get_nth_temp() 6'),
        (4, [33, 34, 35, 38],                   'HdZone._get_nth_temp() 7'),
        (8, [33, 34, 35, 38, 36, 37, 31, 30],   'HdZone._get_nth_temp() 8')
    ])
    def test_get_nth_temp_p2(self, mocker: MockerFixture, count: int, temperatures: List[float], error: str):
        """Positive unit test for HdZone._get_nth_temp() method. It contains the following steps:
            - mock print(), subprocess.run() functions
            - initialize an empty HdZone class
            - ASSERT: if the read temperature (from smartctl) is different from the expected value
        """
        #pylint: disable=line-too-long
        smartctl_output = [
            # SCSI disks
            "smartctl 7.3 2022-02-28 r5338 [x86_64-linux-6.1.0-32-amd64] (local build)\n"
            "Copyright (C) 2002-22, Bruce Allen, Christian Franke, www.smartmontools.org\n"
            "\n"
            "Current Drive Temperature:     XX C\n",

            # SATA SMART attributes 1
            "smartctl 7.3 2022-02-28 r5338 [x86_64-linux-6.1.0-32-amd64] (local build)\n"
            "Copyright (C) 2002-22, Bruce Allen, Christian Franke, www.smartmontools.org\n"
            "\n"
            "190 Airflow_Temperature_Cel 0x0032   075   045   000    Old_age   Always       -       XX\n",

            # SATA SMART attributes 2
            "smartctl 7.3 2022-02-28 r5338 [x86_64-linux-6.1.0-32-amd64] (local build)\n"
            "Copyright (C) 2002-22, Bruce Allen, Christian Franke, www.smartmontools.org\n"
            "\n"
            "194 Temperature_Celsius     0x0002   232   232   000    Old_age   Always       -       XX (Min/Max 17/45)\n"
        ]
        # pylint: enable=line-too-long
        my_td = TestData()
        my_td.create_hd_data(count, temperatures)
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.hwmon_path = my_td.hd_name_list
        my_hdzone.hd_device_names = my_td.hd_name_list
        my_hdzone.hwmon_path = [''] * count
        my_hdzone.sudo=False
        my_hdzone.smartctl_path = '/usr/sbin/smartctl'
        mock_exec_smartclt = MagicMock()
        mock_exec_smartclt.return_value = subprocess.CompletedProcess([], returncode=0, stdout=smartctl_output)
        mocker.patch('smfc.HdZone._exec_smartctl', mock_exec_smartclt)
        for i in range(count):
            s = smartctl_output[random.randint(0, 2)].replace('XX', str(temperatures[i]))
            mock_exec_smartclt.return_value = subprocess.CompletedProcess([], returncode=0, stdout=s)
            assert my_hdzone._get_nth_temp(i) == temperatures[i], error
        del my_hdzone
        del my_td

    @pytest.mark.parametrize("operation, exception, error", [
        # 0. hwmon - FileNotFoundError
        (0, FileNotFoundError,  'HdZone._get_nth_temp() 9'),
        # 1. hwmon - IndexError
        (1, IndexError,         'HdZone._get_nth_temp() 10'),
        # 2. hwmon - ValueError
        (2, ValueError,         'HdZone._get_nth_temp() 11'),
        # 3. smartctl - FileNotFoundError
        (3, FileNotFoundError,  'HdZone._get_nth_temp() 12'),
        # 4. smartctl - IndexError
        (4, IndexError,         'HdZone._get_nth_temp() 13'),
        # 5. smartctl - ValueError
        (5, ValueError,         'HdZone._get_nth_temp() 14')
    ])
    def test_get_nth_temp_n1(self, mocker: MockerFixture, operation: int, exception: Any, error: str):
        """Primitive negative test function. It contains the following steps:
            - mock print(), subprocess.run() functions
            - initialize an empty HdZone class
            - call HdZone._get_nth_temp()
            - ASSERT: if no assertion raised
        """
        index = 0
        my_td = TestData()
        my_td.create_hd_data(1, [32])
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.hwmon_path = my_td.hd_files
        my_hdzone.hd_device_names = my_td.hd_name_list
        my_hdzone.count = 1
        my_hdzone.sudo = False
        # FileNotFoundError: invalid file name with hwmon
        if operation == 0:
            my_hdzone.hwmon_path[0] = '/tmp/non_existent_dir/non_existent_file'
        # IndexError: index error with hwmon
        elif operation == 1:
            index = 1000
        # ValueError: invalid temperature with hwmon
        elif operation == 2:
            os.system('echo "invalid value" >'+my_hdzone.hwmon_path[0])
        # FileNotFoundError: invalid file name with smartctl
        elif operation == 3:
            my_hdzone.hwmon_path[0] = ''
            my_hdzone.smartctl_path = '/tmp/non_existent_dir/non_existent_file'
        # IndexError: index error with smartctl
        elif operation == 4:
            my_hdzone.hwmon_path[0] = ''
            index = 1000
        # ValueError: temperature not found in smartctl's output
        elif operation == 5:
            my_hdzone.hwmon_path[0] = ''
            my_hdzone.smartctl_path = '/usr/sbin/smartctl'
            mock_subprocess_run = MagicMock()
            mock_subprocess_run.return_value = subprocess.CompletedProcess([], returncode=0,
                                                                           stdout='invalid\ninvalid\ninvalid\n')
            mocker.patch('subprocess.run', mock_subprocess_run)
        with pytest.raises(Exception) as cm:
            my_hdzone._get_nth_temp(index)
        assert cm.type == exception, error
        del my_td
    #pylint: enable=protected-access

    @pytest.mark.parametrize("states, result, error", [
        ([True, True, True, True, True, True, True, True],         'SSSSSSSS', 'hz get_standby_state_str 1'),
        ([False, False, False, False, False, False, False, False], 'AAAAAAAA', 'hz get_standby_state_str 2'),
        ([True, False, False, False, False, False, False, False],  'SAAAAAAA', 'hz get_standby_state_str 3'),
        ([False, True, False, False, False, False, False, False],  'ASAAAAAA', 'hz get_standby_state_str 4'),
        ([False, False, True, False, False, False, False, False],  'AASAAAAA', 'hz get_standby_state_str 5'),
        ([False, False, False, True, False, False, False, False],  'AAASAAAA', 'hz get_standby_state_str 6'),
        ([False, False, False, False, True, False, False, False],  'AAAASAAA', 'hz get_standby_state_str 7'),
        ([False, False, False, False, False, True, False, False],  'AAAAASAA', 'hz get_standby_state_str 8'),
        ([False, False, False, False, False, False, True, False],  'AAAAAASA', 'hz get_standby_state_str 9'),
        ([False, False, False, False, False, False, False, True],  'AAAAAAAS', 'hz get_standby_state_str 10')
    ])
    def test_get_standby_state_str(self, states: List[bool], result: str, error: str):
        """Positive unit test for HdZone.get_standby_state_str() method. It contains the following steps:
            - initialize an empty HdZone class
            - calls HdZone.get_standby_state_str()
            - ASSERT: if HdZone.get_standby_state_str() returns different from expected result
        """
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.count = 8
        my_hdzone.standby_array_states = states
        assert my_hdzone.get_standby_state_str() == result, error
        del my_hdzone

    @pytest.mark.parametrize("states, in_standby, error", [
        ([True, True, True, True, True, True, True, True],          8, 'HdZone.check_standby_state() 1'),
        ([False, True, True, True, True, True, True, True],         7, 'HdZone.check_standby_state() 2'),
        ([True, False, True, True, True, True, True, True],         7, 'HdZone.check_standby_state() 3'),
        ([True, True, False, True, True, True, True, True],         7, 'HdZone.check_standby_state() 4'),
        ([True, True, True, False, True, True, True, True],         7, 'HdZone.check_standby_state() 5'),
        ([True, True, True, True, False, True, True, True],         7, 'HdZone.check_standby_state() 6'),
        ([True, True, True, True, True, False, True, True],         7, 'HdZone.check_standby_state() 7'),
        ([True, True, True, True, True, True, False, True],         7, 'HdZone.check_standby_state() 8'),
        ([True, True, True, True, True, True, True, False],         7, 'HdZone.check_standby_state() 9'),
        ([True, False, True, True, True, True, True, False],        6, 'HdZone.check_standby_state() 10'),
        ([True, False, True, True, False, True, True, False],       5, 'HdZone.check_standby_state() 11'),
        ([False, False, True, True, False, True, True, False],      4, 'HdZone.check_standby_state() 12'),
        ([False, False, True, False, False, True, True, False],     3, 'HdZone.check_standby_state() 13'),
        ([False, False, True, False, False, True, False, False],    2, 'HdZone.check_standby_state() 14'),
        ([False, False, False, False, False, True, False, False],   1, 'HdZone.check_standby_state() 15'),
        ([False, False, False, False, False, False, False, False],  0, 'HdZone.check_standby_state() 16')
    ])
    def test_check_standby_state(self, mocker:MockerFixture, states: List[bool], in_standby: int, error: str):
        """Positive unit test for HdZone.check_standby_state() method. It contains the following steps:
            - mock print(), HdZone._exec_smartctl() functions
            - initialize an empty HdZone classes
            - call HdZone.check_standby_state()
            - ASSERT: if result is different from the expected one
        """
        smartctl_output = [

            # Device in STANDBY mode.
            "smartctl 7.2 2020-12-30 r5155 [x86_64-linux-5.10.0-0.bpo.5-amd64] (local build)\n"
            "Copyright (C) 2002-20, Bruce Allen, Christian Franke, www.smartmontools.org\n"
            "\n"
            "Device is in STANDBY mode, exit(2)\n",

            # Device is ACTIVE.
            "smartctl 7.2 2020-12-30 r5155 [x86_64-linux-5.10.0-0.bpo.5-amd64] (local build)\n"
            "Copyright (C) 2002-20, Bruce Allen, Christian Franke, www.smartmontools.org\n"
            "\n"
            "=== START OF INFORMATION SECTION ===\n"
            "Model Family:     Samsung based SSDs\n"
            "Device Model:     Samsung SSD 870 QVO 8TB\n"
            "Serial Number:    S5SSNG0NB01828M\n"
            "LU WWN Device Id: 5 002538 f70b0ee2f\n"
            "Firmware Version: SVQ01B6Q\n"
            "User Capacity:    8,001,563,222,016 bytes [8.00 TB]\n"
            "Sector Size:      512 bytes logical/physical\n"
            "Rotation Rate:    Solid State Device\n"
            "Form Factor:      2.5 inches\n"
            "TRIM Command:     Available, deterministic, zeroed\n"
            "Device is:        In smartctl database [for details use: -P show]\n"
            "ATA Version is:   ACS-4 T13/BSR INCITS 529 revision 5\n"
            "SATA Version is:  SATA 3.3, 6.0 Gb/s (current: 6.0 Gb/s)\n"
            "Local Time is:    Sat May 15 14:26:26 2021 CEST\n"
            "SMART support is: Available - device has SMART capability.\n"
            "SMART support is: Enabled\n"
            "Power mode is:    ACTIVE or IDLE\n"
            "\n"
        ]
        results: List[subprocess.CompletedProcess] = []

        count = 8
        for i in range(count):
            results.append(subprocess.CompletedProcess([], returncode=0, stdout=smartctl_output[0 if states[i] else 1]))
        my_td = TestData()
        my_td.create_hd_data(count)
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.count = count
        my_hdzone.sudo=False
        my_hdzone.hd_device_names = my_td.hd_name_list
        my_hdzone.hwmon_path = [''] * count
        my_hdzone.smartctl_path = '/usr/sbin/smartctl'
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_hdzone.log = my_log
        my_hdzone.standby_array_states = [True] * count
        mock_hdzone_exec_smartclt = MagicMock()
        mock_hdzone_exec_smartclt.side_effect = iter(results)
        mocker.patch('smfc.HdZone._exec_smartctl', mock_hdzone_exec_smartclt)
        assert my_hdzone.check_standby_state() == in_standby, error
        del my_td

    @pytest.mark.parametrize("states, count, error", [
        ([False, False, False, False, False, False, False, False],  8, 'HdZone.go_standby_state() 1'),
        ([True, False, False, False, False, False, False, False],   7, 'HdZone.go_standby_state() 2'),
        ([True, True, False, False, False, False, False, False],    6, 'HdZone.go_standby_state() 3'),
        ([True, True, True, False, False, False, False, False],     5, 'HdZone.go_standby_state() 4'),
        ([True, True, True, True, False, False, False, False],      4, 'HdZone.go_standby_state() 5'),
        ([True, True, True, True, True, False, False, False],       3, 'HdZone.go_standby_state() 6'),
        ([True, True, True, True, True, True, False, False],        2, 'HdZone.go_standby_state() 7'),
        ([True, True, True, True, True, True, True, False],         1, 'HdZone.go_standby_state() 8'),
        ([True, True, True, True, True, True, True, True],          0, 'HdZone.go_standby_state() 9')
    ])
    def test_go_standby_state(self, mocker: MockerFixture, states: List[bool], count: int, error: str):
        """Positive unit test for HdZone.go_standby_state() method. It contains the following steps:
            - mock HdZone._exec_smartctl() function
            - initialize an empty HdZone classes
            - calls HdZone.go_standby_state()
            - ASSERT: if the array state is not in fully standby
        """
        my_td = TestData()
        my_td.create_hd_data(8)
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.count = 8
        my_hdzone.sudo = False
        my_hdzone.hd_device_names = my_td.hd_name_list
        my_hdzone.smartctl_path = '/usr/sbin/smartctl'
        my_hdzone.standby_array_states = states
        mock_exec_smartctl = MagicMock()
        mock_exec_smartctl.return_value = subprocess.CompletedProcess([], returncode=0)
        mocker.patch('smfc.HdZone._exec_smartctl', mock_exec_smartctl)
        my_hdzone.go_standby_state()
        assert mock_exec_smartctl.call_count == count, error
        assert my_hdzone.standby_array_states == [True, True, True, True, True, True, True, True], error
        del my_td

    @pytest.mark.parametrize("old_state, states, new_state, error", [
        # 1. No state changes.
        (False, [False, False, False, False, False, False, False, False],   False, 'HdZone.run_standby_guard() 1'),
        (True,  [True, True, True, True, True, True, True, True],           True,  'HdZone.run_standby_guard() 2'),
        # 2. change from ACTIVE to STANDBY.
        (False, [False, True, False, False, False, False, False, False],    True,  'HdZone.run_standby_guard() 3'),
        (False, [False, True, False, True, False, False, False, False],     True,  'HdZone.run_standby_guard() 4'),
        (False, [True, True, True, True, True, True, True, True],           True,  'HdZone.run_standby_guard() 5'),
        # 3. change from STANDBY to ACTIVE.
        (True,  [False, False, False, False, False, False, False, False],   False, 'HdZone.run_standby_guard() 6'),
        (True,  [True, False, False, True, True, True, True, True],         False, 'HdZone.run_standby_guard() 7')
    ])
    def test_run_standby_guard(self, mocker:MockerFixture, old_state: bool, states: List[bool], new_state: bool,
                               error: str):
        """Positive unit test for HdZone.run_standby_guard() method. It contains the following steps:
            - mock HdZone._exec_smartctl() function
            - initialize a Log and an empty HdZone classes
            - calls HdZone.run_standby_guard()
            - ASSERT: if the expected standby_flags are different
        """
        my_td = TestData()
        my_td.create_hd_data(8)
        my_hdzone = HdZone.__new__(HdZone)
        my_hdzone.count = 8
        my_hdzone.hd_device_names = my_td.hd_name_list
        my_hdzone.smartctl_path = '/usr/sbin/smartctl'
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_hdzone.log = my_log
        my_hdzone.go_standby_state = MagicMock(name='go_standby_state')
        my_hdzone.check_standby_state = MagicMock(name='check_standby_state')
        my_hdzone.check_standby_state.return_value = states.count(True)
        my_hdzone.standby_array_states = states
        my_hdzone.standby_flag = old_state
        my_hdzone.standby_hd_limit = 1
        my_hdzone.standby_change_timestamp = time.monotonic()
        mock_exec_smartctl = MagicMock()
        mock_exec_smartctl.return_value = subprocess.CompletedProcess([], returncode=0)
        mocker.patch('smfc.HdZone._exec_smartctl', mock_exec_smartctl)
        my_hdzone.run_standby_guard()
        assert my_hdzone.standby_flag == new_state, error
        del my_hdzone
        del my_td


# End.



#!/usr/bin/env python3
#
#   test_04_cpuzone.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.CpuZone() class.
#
import os
from configparser import ConfigParser
from typing import List
import pyudev
import pytest
from mock import MagicMock
from pytest_mock import MockerFixture
from smfc import Log, Ipmi, FanController, CpuZone
from .test_00_data import TestData

class TestCpuZone:
    """Unit test class for smfc.CpuZone() class"""

    @pytest.mark.parametrize(
        "count, ipmi_zone, temp_calc, steps, sensitivity, polling, min_temp, max_temp, min_level, max_level, error", [
        (1, '0', FanController.CALC_MIN, 5, 4, 2, 30, 50, 35, 100, 'CpuZone.__init__() 1'),
        (2, '1', FanController.CALC_MIN, 6, 5, 3, 35, 55, 36, 99,  'CpuZone.__init__() 2'),
        (4, '2', FanController.CALC_MIN, 7, 6, 4, 40, 60, 37, 98,  'CpuZone.__init__() 3'),
        (1, '3', FanController.CALC_AVG, 5, 4, 2, 30, 50, 35, 100, 'CpuZone.__init__() 4'),
        (2, '4', FanController.CALC_AVG, 6, 5, 3, 35, 55, 36, 99,  'CpuZone.__init__() 5'),
        (4, '5', FanController.CALC_AVG, 7, 6, 4, 40, 60, 37, 98,  'CpuZone.__init__() 6'),
        (1, '6', FanController.CALC_MAX, 5, 4, 2, 30, 50, 35, 100, 'CpuZone.__init__() 7'),
        (2, '7', FanController.CALC_MAX, 6, 5, 3, 35, 55, 36, 99,  'CpuZone.__init__() 8'),
        (4, '8', FanController.CALC_MAX, 7, 6, 4, 40, 60, 37, 98,  'CpuZone.__init__() 9')
    ])
    def test_init_p1(self, mocker:MockerFixture, count: int, ipmi_zone: str, temp_calc: int, steps: int,
                     sensitivity: float, polling: float, min_temp: float, max_temp: float, min_level: int,
                     max_level: int, error: str):
        """Positive unit test for CpuZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Context.list_devices(), smfc.FanController.get_hwmon_path() functions
            - initialize a Config, Log, Ipmi, and CpuZone classes
            - ASSERT: if the CpuZone class attributes contain different from passed values to __init__
            - delete all instances
        """
        dev_list: List[str]

        my_td = TestData()
        my_td.create_cpu_data(count)
        dev_list = [f'DEV{i}' for i in range(count)]
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_context_list_devices = MagicMock()
        mock_context_list_devices.return_value = dev_list
        mocker.patch('pyudev.Context.list_devices', mock_context_list_devices)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.cpu_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '1',
            CpuZone.CV_CPU_IPMI_ZONE: str(ipmi_zone),
            CpuZone.CV_CPU_ZONE_TEMP_CALC: str(temp_calc),
            CpuZone.CV_CPU_ZONE_STEPS: str(steps),
            CpuZone.CV_CPU_ZONE_SENSITIVITY: str(sensitivity),
            CpuZone.CV_CPU_ZONE_POLLING: str(polling),
            CpuZone.CV_CPU_ZONE_MIN_TEMP: str(min_temp),
            CpuZone.CV_CPU_ZONE_MAX_TEMP: str(max_temp),
            CpuZone.CV_CPU_ZONE_MIN_LEVEL: str(min_level),
            CpuZone.CV_CPU_ZONE_MAX_LEVEL: str(max_level)
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        my_cpuzone = CpuZone(my_log, my_udevc, my_ipmi, my_config)
        assert my_cpuzone.log == my_log, error
        assert my_cpuzone.ipmi == my_ipmi
        assert my_cpuzone.ipmi_zone == [int(s) for s in ipmi_zone.split(',' if ',' in ipmi_zone else ' ')], error
        assert my_cpuzone.name == CpuZone.CS_CPU_ZONE, error
        assert my_cpuzone.count == count, error
        assert my_cpuzone.temp_calc == temp_calc, error
        assert my_cpuzone.steps == steps, error
        assert my_cpuzone.sensitivity == sensitivity, error
        assert my_cpuzone.polling == polling, error
        assert my_cpuzone.min_temp == min_temp, error
        assert my_cpuzone.max_temp == max_temp, error
        assert my_cpuzone.min_level == min_level, error
        assert my_cpuzone.max_level == max_level, error
        assert my_cpuzone.hwmon_path == my_td.cpu_files, error
        del my_td

    @pytest.mark.parametrize("error", [
        ('CpuZone.__init__() 10')
    ])
    def test_init_p2(self, mocker:MockerFixture, error: str):
        """Positive unit test CpuZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Context.list_devices(), smfc.FanController.get_hwmon_path() functions
            - initialize a Config, Log, Ipmi, and CpuZone classes
            - ASSERT: if the CpuZone class attributes contain different from default configuration values
        """
        dev_list: List[str]

        my_td = TestData()
        count = 1
        my_td.create_cpu_data(count)
        dev_list = [f'DEV{i}' for i in range(count)]
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_context_list_devices = MagicMock()
        mock_context_list_devices.return_value = dev_list
        mocker.patch('pyudev.Context.list_devices', mock_context_list_devices)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.cpu_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '1'
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        my_cpuzone = CpuZone(my_log, my_udevc, my_ipmi, my_config)
        assert my_cpuzone.log == my_log, error
        assert my_cpuzone.ipmi == my_ipmi
        assert my_cpuzone.ipmi_zone == [Ipmi.CPU_ZONE], error
        assert my_cpuzone.name == CpuZone.CS_CPU_ZONE, error
        assert my_cpuzone.count == 1, error
        assert my_cpuzone.temp_calc == FanController.CALC_AVG, error
        assert my_cpuzone.steps == 6, error
        assert my_cpuzone.sensitivity == 3.0, error
        assert my_cpuzone.polling == 2, error
        assert my_cpuzone.min_temp == 30, error
        assert my_cpuzone.max_temp == 60, error
        assert my_cpuzone.min_level == 35, error
        assert my_cpuzone.max_level == 100, error
        assert my_cpuzone.hwmon_path == my_td.cpu_files, error
        del my_td

    @pytest.mark.parametrize("error", [
        ('CpuZone.__init__() 11')
    ])
    def test_init_n(self, mocker:MockerFixture, error: str):
        """Negative unit test for CpuZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Context.list_devices(), smfc.FanController.get_hwmon_path() functions
            - initialize a Config, Log, Ipmi, and CpuZone classes
            - ASSERT: if no RuntimeError assertion will be generated due to invalid configuration
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_context_list_devices = MagicMock()
        mock_context_list_devices.return_value = []
        mocker.patch('pyudev.Context.list_devices', mock_context_list_devices)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=None)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '1'
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        with pytest.raises(Exception) as cm:
            CpuZone(my_log, my_udevc, my_ipmi, my_config)
        assert cm.type is RuntimeError, error

    #pylint: disable=protected-access
    @pytest.mark.parametrize("count, index, temperatures, error", [
        (1, 0, [38.5],                   'CpuZone._get_nth_temp() 1'),
        (2, 0, [38.5, 40.5],             'CpuZone._get_nth_temp() 2'),
        (2, 1, [38.5, 40.5],             'CpuZone._get_nth_temp() 3'),
        (4, 0, [38.5, 40.5, 42.5, 44.5], 'CpuZone._get_nth_temp() 4'),
        (4, 1, [38.5, 40.5, 42.5, 44.5], 'CpuZone._get_nth_temp() 5'),
        (4, 2, [38.5, 40.5, 42.5, 44.5], 'CpuZone._get_nth_temp() 6'),
        (4, 3, [38.5, 40.5, 42.5, 44.5], 'CpuZone._get_nth_temp() 7')
    ])
    def test_get_nth_temp_p(self, mocker:MockerFixture, count: int, index: int, temperatures: List[float], error: str):
        """Positive unit test for CpuZone._get_nth_temp() method. It contains the following steps:
            - mock print(), pyudev.Context.list_devices(), smfc.FanController.get_hwmon_path() functions
            - initialize a Config, Log, Ipmi, and CpuZone classes
            - ASSERT: if _get_nth_temp() returns a different from the expected temperature
        """
        dev_list: List[str]

        my_td = TestData()
        my_td.create_cpu_data(count, temperatures)
        dev_list = [f'DEV{i}' for i in range(count)]
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_context_list_devices = MagicMock()
        mock_context_list_devices.return_value = dev_list
        mocker.patch('pyudev.Context.list_devices', mock_context_list_devices)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.cpu_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '1'
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        my_cpuzone = CpuZone(my_log, my_udevc, my_ipmi, my_config)
        assert my_cpuzone._get_nth_temp(index) == temperatures[index], error
        del my_td

    @pytest.mark.parametrize("operation, error", [
        (1, 'CpuZone.__init__() 8'),
        (2, 'CpuZone.__init__() 11'),
        (3, 'CpuZone.__init__() 11')
    ])
    def test_get_nth_temp_n(self, mocker: MockerFixture, operation: int, error: str):
        """Negative unit test for CpuZone._get_nth_temp() method. It contains the following steps:
            - mock print(), pyudev.Context.list_devices(), smfc.FanController.get_hwmon_path() functions
            - initialize a Config, Log, Ipmi, and CpuZone classes
            - ASSERT: if _get_nth_temp() will not raise an exception in different error conditions
        """
        dev_list: List[str]
        index: int

        my_td = TestData()
        my_td.create_cpu_data(1)
        dev_list = ['DEV1']
        index = 0
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_context_list_devices = MagicMock()
        mock_context_list_devices.return_value = dev_list
        mocker.patch('pyudev.Context.list_devices', mock_context_list_devices)
        mock_fancontroller_gethwmonpath = MagicMock(side_effect=my_td.cpu_files)
        mocker.patch('smfc.FanController.get_hwmon_path', mock_fancontroller_gethwmonpath)
        my_config = ConfigParser()
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '1'
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_udevc = pyudev.Context.__new__(pyudev.Context)
        my_cpuzone = CpuZone(my_log, my_udevc, my_ipmi, my_config)
        # Generate FileNotFoundError: delete file
        if operation == 1:
            my_td.delete_file(my_td.cpu_files[0])
        # Generate ValueError: write invalid numeric value to the file
        elif operation == 2:
            os.system('echo "invalid value" >' + my_td.cpu_files[0])
        # Generate IndexError: set index overflow
        else:
            # operation == 3
            index = 100
        with pytest.raises(Exception) as cm:
            my_cpuzone._get_nth_temp(index)
        assert cm.type in [IOError, FileNotFoundError, ValueError, IndexError], error
        del my_td
    # pylint: enable=protected-access


# End.



#
#   hdzone.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.HdZone() class implementation.
#
import subprocess
import time
from configparser import ConfigParser
from typing import List
from pyudev import Context, Devices, DeviceNotFoundByFileError
from smfc.fancontroller import FanController
from smfc.ipmi import Ipmi
from smfc.log import Log


class HdZone(FanController):
    """Class for HD zone fan control."""

    # HdZone specific parameters.
    hd_device_names: List[str]          # Device names of the hard disks (e.g. '/dev/disk/by-id/...').

    # Standby guard specific parameters.
    standby_guard_enabled: bool         # Standby guard feature enabled
    standby_hd_limit: int               # Number of HDs in STANDBY state before the full RAID array will go STANDBY
    smartctl_path: str                  # Path for 'smartctl' command
    standby_flag: bool                  # The actual state of the whole HD array
    standby_change_timestamp: float     # Timestamp of the latest change in STANDBY mode
    standby_array_states: List[bool]    # Standby states of HDs
    sudo: bool                          # Use `sudo` command

    # Constant values for the configuration parameters.
    CS_HD_ZONE: str = 'HD zone'
    CV_HD_ZONE_ENABLED: str = 'enabled'
    CV_HD_IPMI_ZONE: str = 'ipmi_zone'
    CV_HD_ZONE_TEMP_CALC: str = 'temp_calc'
    CV_HD_ZONE_STEPS: str = 'steps'
    CV_HD_ZONE_SENSITIVITY: str = 'sensitivity'
    CV_HD_ZONE_POLLING: str = 'polling'
    CV_HD_ZONE_MIN_TEMP: str = 'min_temp'
    CV_HD_ZONE_MAX_TEMP: str = 'max_temp'
    CV_HD_ZONE_MIN_LEVEL: str = 'min_level'
    CV_HD_ZONE_MAX_LEVEL: str = 'max_level'
    CV_HD_ZONE_HD_NAMES: str = 'hd_names'
    CV_HD_ZONE_SMARTCTL_PATH: str = 'smartctl_path'
    CV_HD_ZONE_STANDBY_GUARD_ENABLED: str = 'standby_guard_enabled'
    CV_HD_ZONE_STANDBY_HD_LIMIT: str = 'standby_hd_limit'


    def __init__(self, log: Log, udevc: Context, ipmi: Ipmi, config: ConfigParser, sudo: bool) -> None:
        """Initialize the HdZone class. Abort in case of configuration errors.

        Args:
            log (Log): reference to a Log class instance
            udevc (Context): reference to an udev database connection (instance of Context from pyudev)
            ipmi (Ipmi): reference to an Ipmi class instance
            config (configparser.ConfigParser): reference to the configuration (default=None)

        Raises:
            ValueError: Parameter `hd_names=` is not specified in the configuration
        """
        hd_names: str   # String for hd_names=
        count: int      # HDD count.

        # Save and validate HdZone class-specific parameters.
        hd_names = config[self.CS_HD_ZONE].get(self.CV_HD_ZONE_HD_NAMES)
        if not hd_names:
            raise ValueError('Parameter hd_names= is not specified.')
        if "\n" in hd_names:
            self.hd_device_names = hd_names.splitlines()
        else:
            self.hd_device_names = hd_names.split()
        # Set count.
        count = len(self.hd_device_names)
        # Save sudo flag.
        self.sudo = sudo

        # Iterate through each disk.
        self.hwmon_path=[]
        for i in range(count):
            # Find a device in udev database based on disk name.
            try:
                block_dev = Devices.from_device_file(udevc, self.hd_device_names[i])
            except DeviceNotFoundByFileError:
                raise ValueError(f'hd_names= parameter error: \'{self.hd_device_names[i]}\' cannot be reached.') \
                    from DeviceNotFoundByFileError
            # Add the hwmon path string for NVME/SATA/HDD disks or '' for SAS/SCSI disks.
            self.hwmon_path.append(self.get_hwmon_path(udevc, block_dev.parent))

        # Save path for `smartctl` command.
        self.smartctl_path = config[HdZone.CS_HD_ZONE].get(HdZone.CV_HD_ZONE_SMARTCTL_PATH, '/usr/sbin/smartctl')

        # Initialize FanController class.
        super().__init__(log, ipmi,
            config[HdZone.CS_HD_ZONE].get(HdZone.CV_HD_IPMI_ZONE, fallback=f'{Ipmi.HD_ZONE}'),
            HdZone.CS_HD_ZONE, count,
            config[HdZone.CS_HD_ZONE].getint(HdZone.CV_HD_ZONE_TEMP_CALC, fallback=FanController.CALC_AVG),
            config[HdZone.CS_HD_ZONE].getint(HdZone.CV_HD_ZONE_STEPS, fallback=4),
            config[HdZone.CS_HD_ZONE].getfloat(HdZone.CV_HD_ZONE_SENSITIVITY, fallback=2),
            config[HdZone.CS_HD_ZONE].getfloat(HdZone.CV_HD_ZONE_POLLING, fallback=10),
            config[HdZone.CS_HD_ZONE].getfloat(HdZone.CV_HD_ZONE_MIN_TEMP, fallback=32),
            config[HdZone.CS_HD_ZONE].getfloat(HdZone.CV_HD_ZONE_MAX_TEMP, fallback=46),
            config[HdZone.CS_HD_ZONE].getint(HdZone.CV_HD_ZONE_MIN_LEVEL, fallback=35),
            config[HdZone.CS_HD_ZONE].getint(HdZone.CV_HD_ZONE_MAX_LEVEL, fallback=100)
        )

        # Read and validate the configuration of standby guard if enabled.
        self.standby_guard_enabled = config[HdZone.CS_HD_ZONE].getboolean(HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED,
                                                                          fallback=False)
        if self.count == 1:
            self.log.msg(Log.LOG_INFO, '   WARNING: Standby guard is disabled ([HD zone] count=1')
            self.standby_guard_enabled = False
        if self.standby_guard_enabled:
            self.standby_array_states = [False] * self.count
            # Read and validate further parameters.
            self.standby_hd_limit = config[HdZone.CS_HD_ZONE].getint(HdZone.CV_HD_ZONE_STANDBY_HD_LIMIT, fallback=1)
            if self.standby_hd_limit < 0:
                raise ValueError('standby_hd_limit < 0')
            if self.standby_hd_limit > self.count:
                raise ValueError('standby_hd_limit > count')
            # Get the current power state of the HD array.
            n = self.check_standby_state()
            # Set calculated parameters.
            self.standby_change_timestamp = time.monotonic()
            self.standby_flag = n == self.count

        # Print configuration in CONFIG log level (or higher).
        if self.log.log_level >= Log.LOG_CONFIG:
            self.log.msg(Log.LOG_CONFIG, f'   {self.CV_HD_ZONE_HD_NAMES} = {self.hd_device_names}')
            self.log.msg(Log.LOG_CONFIG, f'   {self.CV_HD_ZONE_SMARTCTL_PATH} = {self.smartctl_path}')
            if self.standby_guard_enabled:
                self.log.msg(Log.LOG_CONFIG, '   Standby guard is enabled:')
                self.log.msg(Log.LOG_CONFIG, f'     {self.CV_HD_ZONE_STANDBY_HD_LIMIT} = {self.standby_hd_limit}')
            else:
                self.log.msg(Log.LOG_CONFIG, '   Standby guard is disabled')

    def callback_func(self) -> None:
        """Call-back function execute standby guard."""
        if self.standby_guard_enabled:
            self.run_standby_guard()

    def _exec_smartctl(self, arguments: List[str]) -> subprocess.CompletedProcess:
        """Execution of the `smartctl` command.
            Args:
                arguments (List[str]): list of argument of `smartctl` command
            Raises:
                FileNotFoundError: command not found
                RuntimeError: sudo error
        """
        r :   subprocess.CompletedProcess
        args: List[str]

        # Execute `smartctl` command.
        try:
            # Build argument list.
            args = []
            if self.sudo:
                args.append('sudo')
            args.append(self.smartctl_path)
            args.extend(arguments)
            r = subprocess.run(args, check=False, capture_output=True, text=True)
            # In case if sudo return code report execution problem (for smartctl it could be any SMART error)
            if r.returncode != 0 and self.sudo and 'sudo' in r.stderr:
                raise RuntimeError(f'sudo error ({r.returncode}): {r.stderr}!')
        except FileNotFoundError as e:
            raise e
        return r

    def _get_nth_temp(self, index: int) -> float:
        """Get the temperature of the nth element in the hwmon list. This is a specific implementation for HD zone.
        Args:
            index (int): index in hwmon list
        Returns:
            float: temperature value
        Raises:
            FileNotFoundError:  file or command cannot be found
            IOError:            file cannot be opened
            ValueError:         invalid temperature value
            IndexError:         invalid index
        """
        value: float = 100      # Read temperature value.

        # Use 'smartctl' command for reading HD temperature in case of empty HWMON path.
        if not self.hwmon_path[index]:
            r: subprocess.CompletedProcess  # result of the executed process
            output_lines: List[str]         # Lines of the output text.
            line: str                       # One line.
            found: bool                     # Temperature value was found.

            # Read disk temperature with calling `smartctl -a /dev/...` command.
            try:
                r = self._exec_smartctl(['-a', self.hd_device_names[index]])
                # Parse the output of `smartctl` command.
                output_lines = str(r.stdout).splitlines()
                found = False
                for line in output_lines:

                    # SCSI type of temperature reporting, like:
                    # `Current Drive Temperature:     37 C`
                    if 'Current Drive Temperature' in line:
                        value = float(line.split(':')[-1].strip().split()[0])
                        found = True
                        break

                    # pylint: disable=C0301
                    # ATA/SATA type of temperature reporting, like:
                    # `190 Airflow_Temperature_Cel 0x0032   075   045   000    Old_age   Always       -       25`
                    # `194 Temperature_Celsius     0x0002   232   232   000    Old_age   Always       -       28 (Min/Max 17/45)`
                    # pylint: enable=C0301
                    if 'Temperature' in line:
                        s = line.split()
                        value = float(s[9])
                        found = True
                        break

                # If we did not find any matching temperature pattern.
                if not found:
                    raise ValueError('smartctl output: Temperature value cannot be found!')

            except (FileNotFoundError, RuntimeError, ValueError, IndexError) as e:
                raise e

        # Read temperature from a HWMON file.
        else:
            try:
                with open(self.hwmon_path[index], "r", encoding="UTF-8") as f:
                    value = float(f.read()) / 1000
            except (IOError, FileNotFoundError, ValueError, IndexError) as e:
                raise e

        return value

    def get_standby_state_str(self) -> str:
        """Get a string representing the power state of the HD array with a character.
        Returns:
            str:   standby state string where all HD represented with a character (A-ACTIVE, S-STANDBY)
        """
        result: str = ''    # Result string

        for i in range(self.count):
            if self.standby_array_states[i]:
                result += 'S'
            else:
                result += 'A'
        return result

    def check_standby_state(self) -> int:
        """Check the actual power state of the HDs in the array and store them in 'standby_states'.

        Returns:
            int:   number of HDs in STANDBY mode
        """
        r: subprocess.CompletedProcess      # Result of the executed process.

        # Check the current power state of the HDs
        for i in range(self.count):
            self.standby_array_states[i] = False
            r = self._exec_smartctl(['-i', '-n', 'standby', self.hd_device_names[i]])
            if str(r.stdout).find("STANDBY") != -1:
                self.standby_array_states[i] = True
        self.log.msg(Log.LOG_DEBUG, f'Standby guard: current state is {self.get_standby_state_str()}.')
        return self.standby_array_states.count(True)

    def go_standby_state(self):
        """Put active HDs to STANDBY state in the array (based on the actual state of 'standby_states')."""
        # Iterate through HDs list
        for i in range(self.count):
            # if the HD is ACTIVE
            if not self.standby_array_states[i]:
                # then move it to STANDBY state
                self._exec_smartctl(['-s', 'standby,now', self.hd_device_names[i]])
                self.standby_array_states[i] = True

    def run_standby_guard(self):
        """Monitor changes in the power state of an HD array and help them to move to STANDBY state together.
        This feature is implemented in the following way:
            * step 1: Checks the power state of all HDs in array
            * step 2: Check if the array is going to STANDBY (i.e. was ACTIVE before and reached the limit with number
                      of HDs in STANDBY state), put the remaining active members to STANDBY state and log the event
            * step 3: Check if the array is waking up (i.e. was in STANDBY state before and there is any ACTIVE
                      HD(s) in the array) and log the event
        """
        hds_in_standby: int     # HDDs in standby mode
        hours: float            # Elapsed time in minutes
        cur_time: float         # New timestamp for STANDBY change

        # Step 1: check the current power state of the HD array
        hds_in_standby = self.check_standby_state()
        cur_time = time.monotonic()

        # Step 2: check if the array is going to STANDBY state.
        if not self.standby_flag and hds_in_standby >= self.standby_hd_limit:
            hours = (cur_time - self.standby_change_timestamp) / float(3600)
            self.log.msg(Log.LOG_INFO,
                        f'Standby guard: Status change ACTIVE > STANDBY (after {hours:.1f} hours, '
                         f'{self.get_standby_state_str()})')
            self.go_standby_state()
            self.standby_flag = True
            self.standby_change_timestamp = cur_time

        # Step 3: check if the array is waking up.
        elif self.standby_flag and hds_in_standby < self.count:
            hours = (cur_time - self.standby_change_timestamp) / float(3600)
            self.log.msg(Log.LOG_INFO,
                         f'Standby guard: Status change STANDBY > ACTIVE (after {hours:.1f} hours, '
                         f'{self.get_standby_state_str()})')
            self.standby_flag = False
            self.standby_change_timestamp = cur_time


# End.
